'use strict'

const Parser = require('../lib/parser')
const assert = require('assert')

/* global describe it */

const template = `
  <div>
    <input />
    <slot/>
  </div>
`

const script = `
  /**
   * The generic component
   * Sub description
   * 
   * 
   * @public
   * @alpnum azert0 123456789
   * @generic Keyword generic description
   * @multiline Keyword multiline
   *            description
   * @special-char {$[ç(àë£€%µù!,|\`_\\<>/_ç^?;.:/!§)]}
   * @punctuations !,?;.:!
   * @operators -/+<>=*%
   * 
   * @slot inputs - Use this slot to define form inputs ontrols
   * @slot actions - Use this slot to define form action buttons controls
   * @slot footer - Use this slot to define form footer content.
   *
   * @model
   */
  export default {}
  `

describe('Parser', () => {
  describe('constructor(options)', () => {
    it('should successfully create new object', () => {
      const filename = './fixtures/checkbox.vue'
      const defaultMethodVisibility = 'public'
      const options = {
        source: { script, template },
        filename,
        defaultMethodVisibility
      }
      const parser = new Parser(options)

      assert.notEqual(parser.ast, null)
      assert.equal(parser.template, template)
      assert.equal(parser.filename, filename)
      assert.equal(parser.defaultMethodVisibility, defaultMethodVisibility)
    })

    it('should successfully create new object with missing script', () => {
      const options = {
        source: { template }
      }
      const parser = new Parser(options)

      assert.equal(parser.ast, null)
      assert.equal(parser.template, template)
    })

    it('should successfully create new object with empty script', () => {
      const options = {
        source: { template, script: '' }
      }
      const parser = new Parser(options)

      assert.equal(parser.ast, null)
      assert.equal(parser.template, template)
    })
  })

  describe('walk()', () => {
    describe('description', () => {
      it('should successfully emit component description', (done) => {
        const script = `
          /**
           * Component description
           * on multiline
           */
          export default {}
        `
        const options = { source: { script } }
        const parser = new Parser(options)

        parser.walk().on('description', (value) => {
          assert.equal(value, 'Component description on multiline')

          done()
        })
      })
    })

    describe('keywords', () => {
      it('should successfully emit component keywords', (done) => {
        const script = `
          /**
           * @name my-checkbox
           */
          export default {}
        `
        const options = { source: { script } }
        const parser = new Parser(options)

        parser.walk().on('keywords', (value) => {
          assert.deepEqual(value, [ { name: 'name', description: 'my-checkbox' } ])

          done()
        })
      })
    })

    describe('export default expression', () => {
      it('should successfully emit component name', (done) => {
        const script = `
          import child from 'child.vue'

          const component = {
            name: 'hello'
          }

          export default component
        `
        const options = { source: { script } }
        const parser = new Parser(options)

        parser.walk().on('name', (value) => {
          assert.equal(value, 'hello')

          done()
        })
      })

      it('should failed with missing exporting identifier', (done) => {
        const script = `
          export default component
        `
        const options = { source: { script } }
        const parser = new Parser(options)

        parser.walk().on('end', () => done())
      })
    })

    describe('parseTemplate()', () => {
      it('should successfully emit default slot', (done) => {
        const filename = './fixtures/checkbox.vue'
        const defaultMethodVisibility = 'public'
        const options = {
          source: { template },
          filename,
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('slot', (slot) => {
          assert.equal(slot.name, 'default')
          assert.equal(slot.description, null)
          done()
        })
      })

      it('should successfully emit default slot with description', (done) => {
        const filename = './fixtures/checkbox.vue'
        const defaultMethodVisibility = 'public'
        const template = `
          <div>
            <!-- a comment -->
            <p>Hello</p>
            <!-- default slot -->
            <slot/>
          </div>
        `
        const options = {
          source: { template },
          filename,
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('slot', (slot) => {
          assert.equal(slot.name, 'default')
          assert.equal(slot.description, 'default slot')
          done()
        })
      })
    })

    describe('parseComponentName()', () => {
      it('should successfully emit component name with only template', (done) => {
        const filename = './fixtures/checkbox.vue'
        const defaultMethodVisibility = 'private'
        const template = `
          <div>
            <!-- a comment -->
            <p>Hello</p>
          </div>
        `
        const options = {
          source: { template },
          filename,
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('name', (name) => {
          assert.equal(name, 'checkbox')
          done()
        })
      })

      it('should successfully emit component name with explicit name', (done) => {
        const filename = './fixtures/checkbox.vue'
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            name: 'myInput'
          }
        `
        const options = {
          source: { script },
          filename,
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('name', (name) => {
          assert.equal(name, 'my-input')
          done()
        })
      })
    })

    describe('extractProperties(property)', () => {
      it('should successfully emit generic prop', (done) => {
        const filename = './fixtures/checkbox.vue'
        const defaultMethodVisibility = 'public'
        const script = `
          export default {
            props: {
              id: { type: String, default: '$id' }
            }
          }
        `
        const options = {
          source: { script },
          filename,
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('props', (prop) => {
          assert.equal(prop.visibility, 'public')
          assert.equal(prop.name, 'id')
          assert.equal(prop.description, null)
          assert.deepEqual(prop.keywords, [])
          assert.deepEqual(prop.value, { type: 'String', default: '$id' })
          done()
        })
      })

      it('should successfully emit v-model prop', (done) => {
        const filename = './fixtures/checkbox.vue'
        const defaultMethodVisibility = 'public'
        const script = `
          export default {
            props: {
              /**
               * @model v-model keyword
               */
              value: { type: String }
            }
          }
        `
        const options = {
          source: { script },
          filename,
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('props', (prop) => {
          assert.equal(prop.visibility, 'public')
          assert.equal(prop.name, 'v-model')
          assert.equal(prop.description, '')
          assert.deepEqual(prop.keywords, [ { name: 'model', description: 'v-model keyword' } ])
          assert.deepEqual(prop.value, { type: 'String' })
          done()
        })
      })

      it('should successfully emit an unknow item', (done) => {
        const filename = './fixtures/checkbox.vue'
        const defaultMethodVisibility = 'public'
        const script = `
          export default {
            unknow: {
              /**
               * @model v-model keyword
               */
              value: { type: String }
            }
          }
        `
        const options = {
          source: { script },
          filename,
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('unknow', (prop) => {
          done()
        })
      })

      it('should successfully emit methods', (done) => {
        const filename = './fixtures/checkbox.vue'
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            methods: {
              getValue: (ctx) => {}
            }
          }
        `
        const options = {
          source: { script },
          filename,
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('methods', (prop) => {
          assert.equal(prop.visibility, 'private')
          assert.equal(prop.name, 'getValue')
          assert.equal(prop.description, null)
          assert.deepEqual(prop.keywords, [])
          assert.deepEqual(prop.params, [{
            name: 'ctx',
            type: 'Identifier',
            start: 76,
            end: 79,
            range: [76, 79]
          }])
          done()
        })
      })

      it('should emit nothing', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            desc: 'desc-v'
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('end', () => done())
      })
    })

    describe('subWalk(entry)', () => {
      it('should emit event without description', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            desc: () => {
              this.$emit('loading', true)
            }
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('event', (event) => {
          assert.equal(event.name, 'loading')
          assert.equal(event.description, null)
          assert.equal(event.visibility, 'public')
          assert.deepEqual(event.keywords, [])
          done()
        })
      })

      it('should emit event with description', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            desc: () => {
              /**
               * loading event
               *
               * @protected
               */
              this.$emit('loading', true)
            }
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('event', (event) => {
          assert.equal(event.name, 'loading')
          assert.equal(event.description, 'loading event')
          assert.equal(event.visibility, 'protected')
          assert.deepEqual(event.keywords, [ { name: 'protected', description: '' } ])
          done()
        })
      })

      it('should emit event with @event keyword', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            desc () {
              /**
               * Event description
               * 
               * @event loading
               */
              this.$emit(name, true)
            }
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('event', (event) => {
          assert.equal(event.name, 'loading')
          assert.equal(event.description, 'Event description')
          assert.equal(event.visibility, 'public')
          assert.deepEqual(event.keywords, [ { name: 'event', description: 'loading' } ])

          done()
        })
      })

      it('should emit event with missing @event value keyword', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            desc () {
              /**
               * @event
               */
              this.$emit(name)
            }
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('error', (err) => {
          assert.ok(/Missing keyword value for @event/.test(err.message))

          done()
        })
      })

      it('should emit event with identifier name', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            desc: () => {
              const name = 'loading'
              /**
               * loading event
               */
              this.$emit(name, true)
            }
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('event', (event) => {
          assert.equal(event.name, 'loading')
          assert.equal(event.description, 'loading event')
          assert.equal(event.visibility, 'public')
          assert.deepEqual(event.keywords, [])
          done()
        })
      })

      it('should emit event with recursive identifier name', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            desc: () => {
              const pname = 'loading'
              const value = true
              const name = pname
              /**
               * loading event
               * @protected
               */
              this.$emit(name, true)
            }
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('event', (event) => {
          assert.equal(event.name, 'loading')
          assert.equal(event.description, 'loading event')
          assert.equal(event.visibility, 'protected')
          assert.deepEqual(event.keywords, [ { name: 'protected', description: '' } ])
          done()
        })
      })

      it('should emit event with external identifier name', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          const ppname = 'loading'

          export default {
            desc: () => {
              const pname = ppname
              const name = pname
              /**
               * loading event
               */
              this.$emit(name, true)
            }
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('event', (event) => {
          assert.equal(event.name, 'loading')
          assert.equal(event.description, 'loading event')
          assert.equal(event.visibility, 'public')
          assert.deepEqual(event.keywords, [])
          done()
        })
      })

      it('should failed to found identifier name', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            desc: () => {
              const pname = ppname
              const name = pname
              /**
               * loading event
               */
              this.$emit(name, true)
            }
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }
        const parser = new Parser(options)

        parser.walk().on('event', (event) => {
          assert.equal(event.name, '****unhandled-event-name****')
          assert.equal(event.description, 'loading event')
          assert.equal(event.visibility, 'public')
          assert.deepEqual(event.keywords, [])
          done()
        })
      })

      it('should skip already sent event', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            loading: () => {
              this.$emit('loading')
            },
            loading2: () => {
              this.$emit('loading', true)
            }
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }

        const parser = new Parser(options)
        let eventCount = 0

        parser.walk()
          .on('event', (event) => {
            assert.equal(event.name, 'loading')
            assert.equal(event.description, undefined)
            assert.equal(event.visibility, 'public')
            assert.deepEqual(event.keywords, [])

            eventCount++
          })
          .on('end', () => {
            assert.equal(eventCount, 1)

            done()
          })
      })

      it('should skip malformated event emission', (done) => {
        const defaultMethodVisibility = 'private'
        const script = `
          export default {
            loading2: () => {
              this.$emit
            }
          }
        `
        const options = {
          source: { script },
          defaultMethodVisibility
        }

        const parser = new Parser(options)
        let eventCount = 0

        parser.walk()
          .on('event', () => {
            eventCount++
          })
          .on('end', () => {
            assert.equal(eventCount, 0)

            done()
          })
      })
    })
  })
})
